2002-08-22 12:05:11

by David Howells

[permalink] [raw]
Subject: [PATCH] printk simultaneous oops disentangling

Hi Linus,

Here's a patch to stop multiple simultaneous oopses on an SMP system from
interleaving with and overwriting bits of each other. It only permits lock
breaking if the printk lock is held by the same CPU.

David

--- linux-2.5.31/kernel/printk.c Mon Jul 22 11:58:29 2002
+++ linux-trace-2531/kernel/printk.c Thu Aug 22 12:51:24 2002
@@ -390,6 +390,9 @@
logged_chars++;
}

+/* cpu currently holding logbuf_lock */
+static volatile int printk_cpu = -1;
+
/*
* This is printk. It can be called from any context. We want it to work.
*
@@ -412,8 +415,9 @@
static char printk_buf[1024];
static int log_level_unknown = 1;

- if (oops_in_progress) {
- /* If a crash is occurring, make sure we can't deadlock */
+ if (oops_in_progress && printk_cpu==smp_processor_id()) {
+ /* If a crash is occurring during printk() on this CPU,
+ * make sure we can't deadlock */
spin_lock_init(&logbuf_lock);
/* And make sure that we print immediately */
init_MUTEX(&console_sem);
@@ -421,6 +425,7 @@

/* This stops the holder of console_sem just where we want him */
spin_lock_irqsave(&logbuf_lock, flags);
+ printk_cpu = smp_processor_id();

/* Emit the output into the temporary buffer */
va_start(args, fmt);
@@ -450,6 +455,7 @@
* On some architectures, the consoles are not usable
* on secondary CPUs early in the boot process.
*/
+ printk_cpu = -1;
spin_unlock_irqrestore(&logbuf_lock, flags);
goto out;
}
@@ -458,6 +464,7 @@
* We own the drivers. We can drop the spinlock and let
* release_console_sem() print the text
*/
+ printk_cpu = -1;
spin_unlock_irqrestore(&logbuf_lock, flags);
console_may_schedule = 0;
release_console_sem();
@@ -467,6 +474,7 @@
* allows the semaphore holder to proceed and to call the
* console drivers with the output which we just produced.
*/
+ printk_cpu = -1;
spin_unlock_irqrestore(&logbuf_lock, flags);
}
out:


2002-08-22 15:47:06

by Benjamin LaHaise

[permalink] [raw]
Subject: Re: [PATCH] printk simultaneous oops disentangling

On Thu, Aug 22, 2002 at 01:09:18PM +0100, David Howells wrote:
> Hi Linus,
>
> Here's a patch to stop multiple simultaneous oopses on an SMP system from
> interleaving with and overwriting bits of each other. It only permits lock
> breaking if the printk lock is held by the same CPU.

This is still wrong. It should attempt to acquire the locks with a timeout
before trampling on them, as there may be a printk or other console output
in progress on the other cpu.

-ben
--
"You will be reincarnated as a toad; and you will be much happier."

2002-08-22 15:53:21

by Benjamin LaHaise

[permalink] [raw]
Subject: Re: [PATCH] printk simultaneous oops disentangling

On Thu, Aug 22, 2002 at 11:51:14AM -0400, Benjamin LaHaise wrote:
> On Thu, Aug 22, 2002 at 01:09:18PM +0100, David Howells wrote:
> > Hi Linus,
> >
> > Here's a patch to stop multiple simultaneous oopses on an SMP system from
> > interleaving with and overwriting bits of each other. It only permits lock
> > breaking if the printk lock is held by the same CPU.
>
> This is still wrong. It should attempt to acquire the locks with a timeout
> before trampling on them, as there may be a printk or other console output
> in progress on the other cpu.

/me must be having a bad week

The patch is actually right, but bust_spinlocks still blindly stops on locks
that may not need to be stomped on.

-ben
--
"You will be reincarnated as a toad; and you will be much happier."